from .coordinate import Coordinate as __Coordinate
from .difficulty import Difficulty as __Difficulty
from .komoot_image import KomootImage as __KomootImage
from .segment import Segment as __Segment
from .segment_information import SegmentInformation as __SegmentInformation
from .sport_type import SportTypes as __SportTypes
from .surface import Surface as __Surface
from .tour import Tour as __Tour
from .tour_information import TourInformation as __TourInformation
from .tour_summary import TourSummary as __TourSummary
from .way_type import WayType as __WayType
from .waypoint import Waypoint as __Waypoint



from typing import Optional

from pymoot.model.coordinate import Coordinate


class Waypoint:

    def __init__(self, location: Coordinate, index: int, end_index: Optional[int] = None,
                 reference: Optional[str] = None):
        """
        Waypoint object.

        :param location: latitude and longitude of waypoint
        :param index: an index (inclusive) into the tour coordinates array representing the location of the waypoint
        along the tour
        :param end_index: if the waypoint is a highlight segment, the segment ends at this index (exclusive) in the
        tour coordinate array
        :param reference: a namespaced reference to a highlight or OSM POI, e.g. "hl:85124"
        """
        self.location: Coordinate = location
        self.index: int = index
        self.end_index: Optional[int] = end_index
        self.reference: Optional[str] = reference



from typing import Optional


class Segment:

    def __init__(self, segment_type: str, start: int, end: int, reference: Optional[str] = None):
        """
        Segment object.

        :param segment_type: type of segment
        :param start: start of the segment
        :param end: end of the segment
        :param reference: reference of the segment (optional)
        """
        self.segment_type: str = segment_type
        self.start: int = start
        self.end: int = end
        self.reference: Optional[str] = reference


from typing import Dict, Any

from pymoot.model.coordinate import Coordinate

"""
Highlight model description exists here: https://static.komoot.de/doc/external-api/v007/docson/index.html#../schemas/highlight.schema.json
"""


class HighlightTypes:
    HIGHLIGHT_POINT = "highlight_point"
    HIGHLIGHT_SEGMENT = "highlight_segment"


class Highlight:
    def __init__(
            self,
            highlight: Dict[str, Any],
    ):
        """Representation of a highlight."""
        self.id = highlight["id"]
        self.type = highlight["type"]
        self.base_name = highlight["base_name"]
        self.name = highlight["name"]
        self.created_at = highlight["created_at"]
        self.changed_at = highlight["changed_at"]
        self.sport = highlight["sport"]
        self.start_point = Coordinate(
            lat=highlight["start_point"]["lat"],
            lng=highlight["start_point"]["lng"],
            alt=highlight["start_point"]["alt"]
        )
        self.mid_point = Coordinate(
            lat=highlight["mid_point"]["lat"],
            lng=highlight["mid_point"]["lng"],
            alt=highlight["mid_point"]["alt"]
        )
        self.end_point = Coordinate(
            lat=highlight["end_point"]["lat"],
            lng=highlight["end_point"]["lng"],
            alt=highlight["end_point"]["alt"]
        )
        self.distance = highlight["distance"]
        self.elevation_up = highlight["elevation_up"]
        self.elevation_down = highlight["elevation_down"]
        self.score = highlight["score"]
        self.poor_quality = highlight["poor_quality"]
        self.categories = highlight["categories"]
        self.flagged = highlight["flagged"],
        self.front_image = highlight["_embedded"]["front_image"]["src"].split("?")[0] \
            if ("_embedded" in highlight and "front_image" in highlight["_embedded"]
                and highlight["_embedded"]["front_image"] is not None) else None



from typing import Final, List


class WayTypes:
    """
    List of available way types.

    Updated list is available here: https://static.komoot.de/doc/external-api/v007/waytypes.html
    """

    FERRY: Final[str] = "wt#ferry"
    ALPINE_BIKE_D9: Final[str] = "wt#alpine_bike_d9"
    ALPINE_BIKE_D8: Final[str] = "wt#alpine_bike_d8"
    TRAIL_D7: Final[str] = "wt#trail_d7"
    TRAIL_D6: Final[str] = "wt#trail_d6"
    TRAIL_D5: Final[str] = "wt#trail_d5"
    TRAIL_D4: Final[str] = "wt#trail_d4"
    TRAIL_D3: Final[str] = "wt#trail_d3"
    TRAIL_D2: Final[str] = "wt#trail_d2"
    TRAIL_D1: Final[str] = "wt#trail_d1"
    TRAIL: Final[str] = "wt#trail"
    HIKE_D9: Final[str] = "wt#hike_d9"
    HIKE_D8: Final[str] = "wt#hike_d8"
    HIKE_D7: Final[str] = "wt#hike_d7"
    HIKE_D6: Final[str] = "wt#hike_d6"
    HIKE_D5: Final[str] = "wt#hike_d5"
    HIKE_D4: Final[str] = "wt#hike_d4"
    HIKE_D3: Final[str] = "wt#hike_d3"
    HIKE_D2: Final[str] = "wt#hike_d2"
    HIKING_PATH: Final[str] = "wt#hiking_path"
    LONG_HIKING_PATH: Final[str] = "wt#long_hiking_path"
    ALPINE_HIKING_PATH: Final[str] = "wt#alpine_hiking_path"
    MOUNTAIN_HIKING_PATH: Final[str] = "wt#mountain_hiking_path"
    TRACK: Final[str] = "wt#track"
    WAY: Final[str] = "wt#way"
    MINOR_ROAD: Final[str] = "wt#minor_road"
    STREET: Final[str] = "wt#street"
    PRIMARY: Final[str] = "wt#primary"
    SERVICE: Final[str] = "wt#service"
    CYCLEWAY: Final[str] = "wt#cycleway"
    CYCLE_ROUTE: Final[str] = "wt#cycle_route"
    FOOTWAY: Final[str] = "wt#footway"
    MOVABLE_BRIDGE: Final[str] = "wt#movable_bridge"
    UNKNOWN: Final[str] = "wt#unknown"
    OFF_GRID: Final[str] = "wt#off_grid"

    @classmethod
    def list_all(cls) -> List[str]:
        """
        List all possible way types.

        :return: list of supported way types
        """
        return [
            getattr(cls, attr) for attr in dir(cls) if not attr.startswith('__') and not callable(getattr(cls, attr))
        ]


class WayType:

    def __init__(self, way_type: str, amount: float):
        """
        Way type item object.

        :param way_type: type of way
        :param amount: percent (between 0 and 1) of this way type on the tour
        """
        if way_type not in WayTypes.list_all():
            raise ValueError(f"Unexpected way type: {way_type}")
        if not (0 <= amount <= 1):
            raise ValueError("Amount should be between 0 and 1")
        self.way_type: str = way_type
        self.amount: float = amount



from typing import Final, List


class SportTypes:
    """
    List of supported sport types.

    Updated list can be found here: https://static.komoot.de/doc/external-api/v007/sports.html
    """
    # Main sport types:
    HIKING: Final[str] = "hike"
    MOUNTAINEERING: Final[str] = "mountaineering"
    ROAD_CYCLING: Final[str] = "racebike"
    E_ROAD_CYCLING: Final[str] = "e_racebike"
    BIKE_TOURING: Final[str] = "touringbicycle"
    E_BIKE_TOURING: Final[str] = "e_touringbicycle"
    MOUNTAIN_BIKING: Final[str] = "mtb"
    E_MOUNTAIN_BIKING: Final[str] = "e_mtb"
    GRAVEL_RIDING: Final[str] = "mtb_easy"
    E_GRAVEL_RIDING: Final[str] = "e_mtb_easy"
    ENDURO_MOUNTAIN_BIKING: Final[str] = "mtb_advanced"
    E_ENDURO_MOUNTAIN_BIKING: Final[str] = "e_mtb_advanced"
    RUNNING: Final[str] = "jogging"

    # Tracking sport types:
    CLIMBING: Final[str] = "climbing"
    DOWNHILL: Final[str] = "downhillbike"
    CROSS_COUNTRY_SKIING: Final[str] = "nordic"
    NORDIC_WALKING: Final[str] = "nordicwalking"
    SKATING: Final[str] = "skaten"
    ALPINE_SKIING: Final[str] = "skialpin"
    ALPINE_SKI_TOURING: Final[str] = "skitour"
    SLEDDING: Final[str] = "sled"
    SNOWBOARDING: Final[str] = "snowboard"
    SNOW_SHOE: Final[str] = "snowshoe"
    UNICYCLING: Final[str] = "unicycling"
    BIKE: Final[str] = "citybike"
    OTHER: Final[str] = "other"

    @classmethod
    def list_all(cls) -> List[str]:
        """
        List all supported sport types.
        :return: A list of all supported sport types
        """
        return [
            getattr(cls, attr) for attr in dir(cls) if not attr.startswith('__') and not callable(getattr(cls, attr))
        ]



import gzip
import json
import logging
import re
from io import BytesIO

import requests
import dateutil.parser as parser

from typing import Optional, List, Dict, Any, Union

from gpxpy.gpx import GPX

from pymoot.auth import Auth
from pymoot.model.sport_type import SportTypes
from pymoot.model.tour import Tour, TourTypes, TourSortOrders, TourSortFields
from pymoot.url import Url

logger = logging.getLogger("Connector")


class Connector:

    def __init__(self, email: str, password: str):
        """
        Connector to Komoot API.

        :param email: komoot user email
        :param password: komoot user password
        """
        self.auth = Auth(email=email, password=password)
        response = requests.get(Url.ACCOUNT.format(email=email), auth=(email, password))
        if response.status_code == 403:
            raise ConnectionError("Connection to Komoot failed. Please check your credentials")
        rdata = response.json()
        self.auth.set_username(rdata["username"])
        self.auth.set_token(rdata["password"])
        logger.info(f"Logged in as {email}")

    def get_tours(self,
                  limit: Optional[int] = None,
                  page: Optional[int] = None,
                  user_identifier: Optional[str] = None,
                  tour_type: Optional[str] = None,
                  center: Optional[str] = None,
                  max_distance: Optional[int] = None,
                  sport_types: Optional[List[str]] = None,
                  start_date: Optional[str] = None,
                  end_date: Optional[str] = None,
                  tour_name: Optional[str] = None,
                  sort_field: Optional[str] = None,
                  sort_direction: Optional[str] = None) -> List[Tour]:
        """
        Get list of tours from Komoot.

        :param limit: if given, max number of tours to return, else all tours are returned
        :param page: if given, number of page to return (with size given by limit), else return the first page
        :param user_identifier: user identifier to filter tours, if not provided the one of the authenticated user is
        used
        :param tour_type: tour type to return, if not given return all tupes
        :param center: the center of the search area
        :param max_distance: the max distance to the center given above
        :param sport_types: list of sport types to return
        :param start_date: filter tours more recent than this date
        :param end_date: filter tours older than this date
        :param tour_name: tour name to filter by
        :param sort_field: field to sort (default: date)
        :param sort_direction: direction of sort (default: ascending)
        :return: list of Tour objects
        """
        if user_identifier is None:
            user_identifier = self.auth.get_username()
        if tour_type is not None and tour_type not in TourTypes.list_all():
            raise ValueError(f"Unexpected tour type: {tour_type}")
        if center is not None:
            if not re.match(
                pattern=r"^[-+]?\d{1,2}(\.\d+)?,\s*[-+]?\d{1,3}(\.\d+)?$",
                string=center,
            ):
                raise ValueError(
                    f"Unexpected center: {center}. "
                    "Please provide a valid center in the format 'lat, lng' (e.g. '52.520008, 13.404954')."
                )
        if max_distance is None and center is not None:
            raise ValueError("Max distance is required if center is provided")
        if max_distance is not None and center is None:
            raise ValueError("Center is required if max distance is provided")
        if sport_types is not None:
            if not isinstance(sport_types, list):
                raise TypeError("sport_types must be a list")
            for sport_type in sport_types:
                if not isinstance(sport_type, str):
                    raise TypeError("Each sport type must be a string")
                if sport_type not in SportTypes.list_all():
                    raise ValueError(f"Unexpected sport type: {sport_type}")
        if start_date is not None:
            start_date = parser.parse(start_date)
        if end_date is not None:
            end_date = parser.parse(end_date)
        if start_date is not None and end_date is not None and start_date > end_date:
            raise ValueError("Start date must be before end date")
        if sort_field is not None and sort_field not in TourSortFields.list_all():
            raise ValueError(f"Sort on field {sort_field} is not allowed")
        if sort_field == TourSortFields.PROXIMITY and center is None:
            raise ValueError("To sort by proximity, you must provide a center")
        if sort_direction is not None and sort_direction not in TourSortOrders.list_all():
            raise ValueError(f"Unexpected sort_order value: {sort_direction}")

        query = {}
        if limit is not None:
            query["limit"] = limit
        if page is not None:
            query["page"] = page
        if tour_type is not None:
            query["type"] = tour_type
        if center is not None:
            query["center"] = center
        if max_distance is not None:
            query["max_distance"] = max_distance
        if sport_types is not None:
            query["sport_types"] = sport_types
        if start_date is not None:
            query["start_date"] = start_date
        if end_date is not None:
            query["end_date"] = end_date
        if tour_name is not None:
            query["name"] = tour_name
        if sort_direction is not None:
            query["sort_direction"] = sort_direction
        if sort_field is not None:
            query["sort_field"] = sort_field

        has_more = True
        current_page = 0
        tours = []
        while has_more:
            query["page"] = current_page
            response = self._get_page_of_tours(query=query, user_identifier=user_identifier)
            data = response.json()
            tour_list = data["_embedded"]
            tours.extend(tour_list["tours"])
            max_page = data["page"]["totalPages"]
            current_page = data["page"]["number"] + 1
            has_more = (current_page < max_page) if limit is None else False
        return [Tour(tour_dict, authentication=self.auth) for tour_dict in tours]

    def get_tour(self, tour_id: str, as_gpx=False) -> Union[Tour, GPX]:
        """
        Get a tour.

        :param tour_id: tour identifier
        :param as_gpx: if True, get GPX instead of the tour object
        :return:
        """
        return Tour.retrieve_tour_by_id(authentication=self.auth, tour_id=tour_id, as_gpx=as_gpx)

    def update_tour_title(
            self,
            tour_identifier: str,
            new_title: str
    ):
        """
        Get a tour by its ID.
        :param tour_identifier: The ID of the tour
        :param new_title: The new title of the tour
        """

        try:
            response = requests.patch(
                url=Url.ONE_TOUR.format(tour_identifier=tour_identifier),
                auth=(self.auth.get_email(), self.auth.get_password()),
                headers={"content-encoding": "gzip", "content-type": "application/hal+json"},
                data=self._zip_payload(json.dumps({"name": new_title}))
            )
            if response.status_code == 403:
                raise ConnectionError(
                    'Connection to Komoot API failed. Please check your credentials.'
                )
            if response.status_code == 404:
                raise ValueError(f'Invalid tour identifier provided: {tour_identifier}. '
                                 f'Please provide a valid tour identifier.')
            if response.status_code == 500:
                raise ConnectionError(
                    'Internal Server Error. if you requested a FIT file, '
                    'please try again later or try fetching another format.'
                )
        except requests.exceptions.ConnectionError:
            raise ConnectionError(
                'Connection to Komoot API failed. Please check your internet connection.'
            )
        resp = json.loads(response.content.decode('utf-8'))
        return Tour(resp, self.auth)

    def _get_page_of_tours(self, query: Dict[str, Any], user_identifier: str) -> requests.Response:
        """
        Get a page of tours.

        :param query: query arguments to add to the request
        :param user_identifier: the user identifier
        :return:
        """
        response = requests.get(
            url=Url.LIST_TOURS.format(user_identifier=user_identifier),
            auth=(self.auth.get_email(), self.auth.get_password()),
            params=query
        )
        if response.status_code == 403:
            raise ConnectionError("Connection to komoot field: invalid authentication")
        return response

    @staticmethod
    def _zip_payload(payload: str) -> bytes:
        btsio = BytesIO()
        g = gzip.GzipFile(fileobj=btsio, mode='w')
        g.write(bytes(payload, 'utf8'))
        g.close()
        return btsio.getvalue()



import os
import unittest
from unittest.mock import patch, MagicMock

from pymoot.connector import Connector
from pymoot.model.tour import Tour
from tests.mock import mock_response


class TestConnector(unittest.TestCase):

    @classmethod
    @patch("requests.get")
    def setUpClass(cls, mock_get: MagicMock):
        cls.email = "test@example.com"
        cls.password = "password"
        cls.valid_id = "123456"
        cls.invalid_id = "invalid"
        mock_response(mock_req=mock_get,
                      status_code=200,
                      json_response=f"{os.path.dirname(os.path.realpath(__file__))}/resources"
                                    f"/authentication_response.json")
        cls.connector = Connector(email="test@example.com", password="password")

    @patch("requests.get")
    def test_initialization(self, mock_get: MagicMock):
        mock_response(
            mock_req=mock_get,
            status_code=200,
            json_response=f"{os.path.dirname(os.path.realpath(__file__))}/resources/authentication_response.json",
        )
        connector = Connector(
            email=self.email,
            password=self.password,
        )
        self.assertIsInstance(connector, Connector)

    @patch("requests.get")
    def test_get_tours(self, mock_get):
        mock_response(
            mock_req=mock_get,
            status_code=200,
            json_response=f"{os.path.dirname(os.path.realpath(__file__))}/resources/list_tours_response.json",
        )

        tours = self.connector.get_tours(limit=5)
        self.assertIsInstance(tours, list)
        for tour in tours:
            self.assertIsInstance(tour, Tour)
        self.assertEqual(5, len(tours))

    @patch("requests.get")
    def test_get_tour_by_valid_id(self, mock_get):
        mock_response(
            mock_req=mock_get,
            status_code=200,
            json_response=f"{os.path.dirname(os.path.realpath(__file__))}/resources/get_tour_by_id_response.json",
        )
        tour = self.connector.get_tour(tour_id=self.valid_id)
        self.assertIsInstance(tour, Tour)



class SegmentInformation:

    def __init__(self, from_index: int, to_index: int):
        """
        Segment information object.
        :param from_index: start index point (inclusive)
        :param to_index: stop index point (exclusive)
        """
        self.from_index: int = from_index
        self.to_index: int = to_index


from typing import Optional


class Auth:
    def __init__(self, email: str, password: str):
        """
        Authentication object containing credentials.
        :param email: komoot user email
        :param password: komoot user password
        """
        if not email:
            raise ValueError("Username is required")
        if not password:
            raise ValueError("Password is required")
        self._email: str = email
        self._password: str = password
        self._token: Optional[str] = None
        self._username: Optional[str] = None

    def get_email(self) -> str:
        return self._email

    def get_password(self) -> str:
        return self._password

    def get_token(self) -> str:
        return self._token

    def get_username(self) -> str:
        return self._username

    def set_token(self, token: str) -> None:
        self._token = token

    def set_username(self, username: str) -> None:
        self._username = username

    def __str__(self):
        return f"""
        Authentication object:
        E-mail: {self._email}
        Password: {len(self._password) * "*"}
        Username: {self._username[:2]}...{self._username[-1:]}
        Token: {len(self._token) * "*" if self._token else None}
        """



from typing import List, Final


class SurfaceTypes:
    """
    List of available surfaces.

    Updated list can be found here: https://static.komoot.de/doc/external-api/v007/surfaces.html
    """

    # Bike:
    BIKE_ASPHALT: Final[str] = "sb#asphalt"
    BIKE_CONCRETE: Final[str] = "sb#concrete"
    BIKE_PAVED: Final[str] = "sb#paved"
    BIKE_PAVING_STONES: Final[str] = "sb#paving_stones"
    BIKE_COBBLESTONES: Final[str] = "sb#cobblestone"
    BIKE_COBBLES: Final[str] = "sb#cobbles"
    BIKE_COMPACTED_GRAVEL: Final[str] = "sb#compacted"
    BIKE_GRAVEL: Final[str] = "sb#gravel"
    BIKE_GRASS_PAVER: Final[str] = "sb#grass_paver"
    BIKE_WOOD: Final[str] = "sb#wood"
    BIKE_SAND: Final[str] = "sb#sand"
    BIKE_GROUND: Final[str] = "sb#ground"
    BIKE_STONE: Final[str] = "sb#stone"
    BIKE_UNPAVED: Final[str] = "sb#unpaved"
    BIKE_ALPINE: Final[str] = "sb#alpin"
    BIKE_UNKNOWN: Final[str] = "sb#unknown"

    # Foot:
    FOOT_ASPHALT: Final[str] = "sf#asphalt"
    FOOT_CONCRETE: Final[str] = "sf#concrete"
    FOOT_COBBLESTONES: Final[str] = "sf#cobblestone"
    FOOT_PAVING_STONES: Final[str] = "sf#paving_stones"
    FOOT_PAVED: Final[str] = "sf#paved"
    FOOT_COMPACTED_GRAVEL: Final[str] = "sf#compacted"
    FOOT_GRAVEL: Final[str] = "sf#gravel"
    FOOT_GRASS_PAVER: Final[str] = "sf#grass_paver"
    FOOT_WOOD: Final[str] = "sf#wood"
    FOOT_UNPAVED: Final[str] = "sf#unpaved"
    FOOT_SAND: Final[str] = "sf#sand"
    FOOT_GROUND: Final[str] = "sf#ground"
    FOOT_STONE: Final[str] = "sf#stone"
    FOOT_NATURAL: Final[str] = "sf#nature"
    FOOT_ALPINE: Final[str] = "sf#alpin"
    FOOT_UNKNOWN: Final[str] = "sf#unknown"

    # MTB:
    MTB_ASPHALT: Final[str] = "sm#asphalt"
    MTB_CONCRETE: Final[str] = "sm#concrete"
    MTB_COBBLESTONES: Final[str] = "sm#cobblestone"
    MTB_PAVING_STONES: Final[str] = "sm#paving_stones"
    MTB_PAVED: Final[str] = "sm#paved"
    MTB_COMPACTED_GRAVEL: Final[str] = "sm#compacted"
    MTB_GRAVEL: Final[str] = "sm#gravel"
    MTB_GRASS_PAVER: Final[str] = "sm#grass_paver"
    MTB_WOOD: Final[str] = "sm#wood"
    MTB_UNPAVED: Final[str] = "sm#unpaved"
    MTB_SAND: Final[str] = "sm#sand"
    MTB_GROUND: Final[str] = "sm#ground"
    MTB_STONE: Final[str] = "sm#stone"
    MTB_NATURAL: Final[str] = "sm#nature"
    MTB_ALPINE: Final[str] = "sm#alpin"
    MTB_UNKNOWN: Final[str] = "sm#unknown"

    @classmethod
    def list_all(cls) -> List[str]:
        """
        List all possible surfaces.

        :return: list of supported surfaces
        """
        return [
            getattr(cls, attr) for attr in dir(cls) if not attr.startswith('__') and not callable(getattr(cls, attr))
        ]


class Surface:

    def __init__(self, surface_type: str, amount: float):
        """
        Surface item object.

        :param surface_type: type of the surface
        :param amount: percent (between 0 and 1) of this surface on the tour
        """
        if surface_type not in SurfaceTypes.list_all():
            raise ValueError(f"Unexpected surface type: {surface_type}")
        if not (0 <= amount <= 1):
            raise ValueError("Amount should be between 0 and 1")
        self.surface_type: str = surface_type
        self.amount: float = amount



from typing import Final


class Url:
    ACCOUNT: Final[str] = "https://api.komoot.de/v006/account/email/{email}/"
    LIST_TOURS: Final[str] = "https://api.komoot.de/v007/users/{user_identifier}/tours/"
    ONE_TOUR: Final[str] = "https://api.komoot.de/v007/tours/{tour_identifier}"
    HIGHLIGHT_URL: Final[str] = 'https://api.komoot.de/v007/highlights/{highlight_identifier}'



import logging
from io import BytesIO

import requests

from typing import Optional
from PIL import Image


class KomootImage:

    def __init__(self, url: str, templated: bool, client_hash: Optional[str] = None, attribution: Optional[str] = None,
                 attribution_url: Optional[str] = None, media_type: Optional[str] = None):
        """
        Komoot Image object.

        :param url: link to the image
        :param templated: wether the link contains template variables
        :param client_hash: a hash set by the client on upload, useful for de-duplicating images
        :param attribution: if set, clients must show this attribution with the image
        :param attribution_url: if attribution exists, this field can contain a link to the attribution source
        :param media_type: media type of the resource, e.g. image/*
        """
        self.url: str = url,
        self.templated: bool = templated
        self.client_hash: Optional[str] = client_hash
        self.attribution: Optional[str] = attribution
        self.attribution_url: Optional[str] = attribution_url
        self.media_type: Optional[str] = media_type
        self.image: Optional[Image] = None

    def load_image(self):
        """Load image from the url."""
        response = requests.get(self.url)
        response.raise_for_status()
        self.image = Image.open(BytesIO(response.content))



from typing import Optional


class Coordinate:
    def __init__(self, lat: float, lng: float, alt: Optional[float] = None, time: Optional[float] = None):
        """
        Coordinate object.

        :param lat: latitude
        :param lng: longitude
        :param alt: altitude in meters
        :param time: time, must be >=0 to precedent coordinate
        """
        # Check:
        self.check_lat(lat)
        self.check_lng(lng)
        self.check_alt(alt)
        self.check_time(time)

        # Assign:
        self.lat: float = lat
        self.lng: float = lng
        self.alt: Optional[float] = alt
        self.time: Optional[float] = time

    @staticmethod
    def check_lat(lat: float):
        if not -90 <= lat <= 90:
            raise ValueError(f"Invalid latitude: {lat}. Latitude must be between -90 and 90.")

    @staticmethod
    def check_lng(lng: float):
        if not -180 <= lng <= 180:
            raise ValueError(f"Invalid longitude: {lng}. Longitude mst be between -180 and 180.")

    @staticmethod
    def check_alt(alt: float):
        if alt is not None and not -1000 <= alt <= 10000:
            raise ValueError(
                f"Invalid altitude: {alt}. Altitude must be between -1000 and 10000.")

    @staticmethod
    def check_time(time: float):
        if time is not None and time < 0:
            raise ValueError(f"Invalid time: {time}. Time must be equal or above 0.")



from typing import Final, List


class DifficultyGrade:
    """List of available difficulty grades."""

    EASY: Final[str] = "easy"
    MODERATE: Final[str] = "moderate"
    DIFFICULT: Final[str] = "difficult"

    @classmethod
    def list_all(cls) -> List[str]:
        """
        List all possible difficulty grades.

        :return: list of supported difficulty grades
        """
        return [
            getattr(cls, attr) for attr in dir(cls) if not attr.startswith('__') and not callable(getattr(cls, attr))
        ]


class Difficulty:

    def __init__(self, grade: str, explanation_technical: str, explanation_fitness: str):
        """
        Difficulty: describes estimated difficulty of the tour.

        :param grade: difficulty grade
        :param explanation_technical: technical explanation of the difficulty
        :param explanation_fitness: physical explanation of the difficulty
        """
        if grade not in DifficultyGrade.list_all():
            raise ValueError(f"Unexpected difficulty grade: {grade}")
        self.grade: str = grade
        self.explanation_technical: str = explanation_technical
        self.explanation_fitness: str = explanation_fitness



from typing import List

from pymoot.model.segment_information import SegmentInformation


class TourInformation:

    def __init__(self, ti_type: str, segments: List[SegmentInformation]):
        """
        Tour information object
        :param ti_type: type of tour information
        :param segments: list of segment information
        """
        self.ti_type: str = ti_type
        self.segments: List[SegmentInformation] = segments



from typing import List

from pymoot.model.surface import Surface
from pymoot.model.way_type import WayType


class TourSummary:

    def __init__(self, surfaces: List[Surface], way_types: List[WayType]):
        """
        Tour summary object.

        :param surfaces: list of surfaces
        :param way_types: list of way types
        """
        self.surfaces: List[Surface] = surfaces
        self.way_types: List[WayType] = way_types



import logging
import re
from datetime import datetime
from typing import Dict, Any, Optional, List, Final, Union, Self

import gpxpy
import requests
from dateutil import parser
from gpxpy.gpx import GPX

from pymoot.auth import Auth
from pymoot.model.coordinate import Coordinate
from pymoot.model.difficulty import Difficulty
from pymoot.model.highlight import Highlight
from pymoot.model.komoot_image import KomootImage
from pymoot.model.segment import Segment
from pymoot.model.segment_information import SegmentInformation
from pymoot.model.sport_type import SportTypes
from pymoot.model.surface import Surface
from pymoot.model.tour_information import TourInformation
from pymoot.model.tour_summary import TourSummary
from pymoot.model.way_type import WayType
from pymoot.model.waypoint import Waypoint
from pymoot.url import Url

"""
Tour model description exists here: https://static.komoot.de/doc/external-api/v007/docson/index.html#../schemas/users_tours.schema.json
"""


class TourTypes:
    """Possible types of tours."""

    PLANNED: Final[str] = "tour_planned"
    RECORDED: Final[str] = "tour_recorded"

    @classmethod
    def list_all(cls) -> List[str]:
        """
        List all possible tour types.

        :return: list of supported tour types
        """
        return [
            getattr(cls, attr) for attr in dir(cls) if not attr.startswith('__') and not callable(getattr(cls, attr))
        ]


class TourSortOrders:
    """Possible tour sort directions."""

    ASCENDING: Final[str] = "asc"
    DESCENDING: Final[str] = "desc"

    @classmethod
    def list_all(cls) -> List[str]:
        """
        List all possible tour sort orders.

        :return: list of supported tour sort orders
        """
        return [
            getattr(cls, attr) for attr in dir(cls) if not attr.startswith('__') and not callable(getattr(cls, attr))
        ]


class TourSortFields:
    """
    List of tour fields allowed for sort.
    """
    DATE: Final[str] = "date"
    DURATION: Final[str] = "duration"
    ELEVATION: Final[str] = "elevation"
    NAME: Final[str] = "name"
    PROXIMITY: Final[str] = "proximity"

    @classmethod
    def list_all(cls) -> List[str]:
        """
        List all possible sort fields.
        :return: A list of all supported sort fields
        """
        return [
            getattr(cls, attr) for attr in dir(cls) if not attr.startswith('__') and not callable(getattr(cls, attr))
        ]


class Tour:
    """Tour model."""
    def __init__(self, tour: Dict[str, Any], authentication: Auth):
        """
        Tour model.

        Tour model is described like that:
        - id (int): id of the tour
        - type (tour_planned, tour_recorded): type of the tour
        - status (private, public, friends): privacy status of the tour
        - source (str): source of the tour
        - date (int): timestamp when the tour recording started (for recorded tours) or when the tour was created
                      (for planned tours). ISO 8601 with milliseconds required
        - changed_at (int): timestamp when the last update of the tour occurred. ISO 8601 with milliseconds required
        - name (str) : name of the tour
        - kcal_active: effort estimation in kilocalories (active portion)
        - kcal_resting: effort estimation in kilocalories (resting portion)
        - start_point (Coordinate: first coordinate of the tour (with lat, lng and alt)
        - distance (int): 3D length of the tour, in meters
        - duration (int): Total duration from tour start to finish, in seconds - includes pauses
        - elevation_up (int): total elevation gain in meters
        - elevaton_down (int): total elevation loss in meters
        - sport (str): tour sport type
        - vector_map_image: dynamically rendered map with tour line
        - vector_map_image_preview: dynamically rendered map with tour line, proportioned for thumbnails
        - time_in_motion (int): estimation of actual time in motion (for recorded tours only)
        - constitution (int): constitution value used during planning (for planned tours only)
        - query (str): the query string of the tour (for planned tours only)
        - path (waipoints[]): list of waypoints used for planning the tour (for planned tours only)
        - segments (segment[]): describe the kind of the segments. A segment is the way in between two waypoints
                                (two elements of the path). Consequently the amount of path elements is always one
                                larger than the amount segments
        - tour_information: tour information dict containing:
          * type (FERRY|BICYCLE_DISMOUNT|MOVABLE_BRIDGE|UNSUITABLE|RESTRICTED|OFF_GRID|HEAVY_TRAFFIC|PRIVATE_LAND) :
              indicates type of information. e.g. OFF_GRID that the tour contains an off-grid segment. Clients must
              ignore unknown types
          * segments (list of segments informations)
        - summary: tour summary:
          * surfaces: list of surfaces with amount
          * way_types: list of way types with amount
        - difficulty: difficulty of the tour:
          * grade (difficult|moderate|easy)
          * explanation_technical (str): technical difficulty
          * explanation_fitness (str): physical difficulty
        - master_share_url (str): the web URL to this tour's canonical master tour

        :param tour: the tour as dict
        """
        self.auth = authentication
        self.id: str = tour["id"]
        self.type: str = tour["type"]
        self.status: str = tour["status"]
        self.source: str = tour["source"]
        self.date: datetime = parser.parse(tour["date"])
        self.changed_at: datetime = parser.parse(tour["changed_at"])
        self.name: str = tour["name"]
        self.kcal_active: float = tour["kcal_active"]
        self.kcal_resting: float = tour["kcal_resting"]
        self.start_point: Coordinate = Coordinate(lat=tour["start_point"]["lat"], lng=tour["start_point"]["lng"],
                                      alt=tour["start_point"]["alt"], time=None)
        self.distance: float = tour["distance"]
        self.duration: float = tour["duration"]
        self.elevation_up: float = tour["elevation_up"]
        self.elevation_down: float = tour["elevation_down"]
        if tour["sport"] not in SportTypes.list_all():
            raise ValueError(f"Unexpected sport type: {tour['sport']}")
        self.sport: str = tour["sport"]
        if "vector_map_image" in tour:
            self.vector_map_image = KomootImage(
                url=tour["vector_map_image"]["src"],
                templated=tour["vector_map_image"]["templated"] if "templated" in tour["vector_map_image"] else None,
                client_hash=tour["vector_map_image"]["client_hash"] if "client_hash" in tour["vector_map_image"]
                else None,
                attribution=tour["vector_map_image"]["attribution"] if "attribution" in tour["vector_map_image"]
                else None,
                attribution_url=tour["vector_map_image"]["attribution_url"] if "attribution_url"
                                                                               in tour["vector_map_image"] else None,
                media_type=tour["vector_map_image"]["type"] if "type" in tour["vector_map_image"] else None
            )
        else:
            self.vector_map_image = None
            logging.warning("No vector map image found")
        self.time_in_motion: Optional[int] = tour["time_in_motion"] if "time_in_motion" in tour else None
        self.constitution: Optional[int] = tour["constitution"] if "constition" in tour else None
        self.query: Optional[str] = tour["query"] if "query" in tour else None
        self.path: Optional[List[Waypoint]] = self._build_waypoints_list(tour["path"]) if "path" in tour else None
        self.segments: Optional[List[Segment]] = self._build_segments_list(tour["segments"]) \
            if "segments" in tour else None
        self.tour_information = self._build_tour_information_list(tour["tour_information"]) \
            if "tour_information" in tour else None
        self.summary = self._build_summary(tour["summary"]) if "summary" in tour else None
        if "difficulty" in tour:
            difficulty = tour["difficulty"]
            self.difficulty = Difficulty(grade=difficulty["grade"],
                                         explanation_technical=tour["difficulty"]["explanation_technical"],
                                         explanation_fitness=tour["difficulty"]["explanation_fitness"])
        self.master_share_url: str = tour["master_share_url"] if "master_share_url" in tour else None
        self.links_dict = tour['_links'] if '_links' in tour else None
        self.coordinates_link = None
        if self.links_dict is not None and "coordinates_link" in self.links_dict:
            self.coordinates_link = self.links_dict["coordinates"]["href"]
        self.coordinates: List[Coordinate] = []
        self.gpx_track: Optional[GPX] = None

        

    @staticmethod
    def _build_waypoints_list(path: List[Dict[str, Any]]) -> List[Waypoint]:
        """
        Build the list of waypoints from a path.

        :param path: the source path
        :return: list of waypoints
        """
        waypoints = []
        for waypoint in path:
            waypoints.append(Waypoint(
                location=Coordinate(lat=waypoint["location"]["lat"], lng=waypoint["location"]["lng"]),
                index=waypoint["index"],
                end_index=waypoint["end_index"] if "end_index" in waypoint else None,
                reference=waypoint["reference"] if "reference" in waypoint else None
            ))
        return waypoints

    @staticmethod
    def _build_segments_list(segments: List[Dict[str, Any]]) -> List[Segment]:
        """
        Build the list of segments from tour segments.

        :param segments: tour segments as list of dicts
        :return: final list of segments objects
        """
        final_segments = []
        for segment in segments:
            final_segments.append(Segment(
                segment_type=segment["type"],
                start=segment["from"],
                end=segment["to"],
                reference=segment["reference"] if "reference" in segment else None
            ))
        return final_segments

    @staticmethod
    def _build_tour_information_list(tour_information: List[Dict[str, Any]]) -> List[TourInformation]:
        """
        Build a list of tour information objects from tour information list.
        :param tour_information: list of tour information from API
        :return: list of TourInformation objects
        """
        tour_information_list = []
        for ti in tour_information:
            tour_information_list.append(
                TourInformation(
                    ti_type=ti["type"],
                    segments=[SegmentInformation(from_index=segment["from"], to_index=segment["to"])
                              for segment in ti["segments"]]
                )
            )
        return tour_information_list

    @staticmethod
    def _build_summary(summary_dict: Dict[str, Any]) -> TourSummary:
        """
        Build tour summary.
        :param summary_dict: summary dict from API
        :return: TourSummary object
        """
        return TourSummary(
            surfaces=[
                Surface(surface_type=surface["type"], amount=surface["amount"]) for surface in summary_dict["surfaces"]
            ],
            way_types=[
                WayType(way_type=way_type["type"], amount=way_type["amount"]) for way_type in summary_dict["way_types"]
            ]
        )

    def retrieve_coordinates(self) -> Optional[List[Coordinate]]:
        """
        Retrieve coordinates of the tour.

        :param authentication: authentication object
        :return: Coordinates, if any
        """
        if self.coordinates_link is not None:
            response = requests.get(
                url=self.coordinates_link,
                auth=(self.auth.get_email(), self.auth.get_password())
            ).json()["items"]
        else:
            logging.warning("No coordinates link exists for this tour")
            return None

        self.coordinates = [
            Coordinate(
                lat=coordinate["lat"],
                lng=coordinate["lng"],
                alt=coordinate["alt"] if "alt" in coordinate else None,
                time=coordinate["time"] if "time" in coordinate else None
            ) for coordinate in response
        ]
        return self.coordinates

    @classmethod
    def retrieve_tour_by_id(cls, tour_id: str, authentication: Auth, as_gpx: bool) -> Union[Self, GPX]:
        url_append = ""
        if as_gpx:
            url_append = ".gpx"
        response = requests.get(
            url=Url.ONE_TOUR.format(tour_identifier=tour_id) + url_append,
            auth=(authentication.get_email(), authentication.get_password()),
            params={"Type": "application/hal+json"}
        )
        if response.status_code == 403:
            raise ConnectionError("Connection to komoot field: invalid authentication")
        if response.status_code == 404:
            raise ValueError(f"No tour exists with id {tour_id}")
        if as_gpx:
            return gpxpy.parse(response.content)
        else:
            return Tour(response.json(), authentication=authentication)

    def retrieve_gpx(self) -> GPX:
        """
        Retrieve gpx data of a tour.

        :param authentication: authentication object
        :return: GPX data, if any
        """
        self.gpx_track = self.retrieve_tour_by_id(tour_id=self.id, authentication=self.auth, as_gpx=True)
        return self.gpx_track

    def retrieve_tour_highlights(self):
        """Retrieve tour highlights."""
        highlights = []
        for path in self.path:
            if path.reference is not None and (match := re.match(r"hl(p|s):(\d+)", path.reference)):
                highlight_id = match.group(2)
                response = requests.get(
                    url=Url.HIGHLIGHT_URL.format(highlight_identifier=highlight_id),
                    auth=(self.auth.get_email(), self.auth.get_password()),
                    headers={"content-encoding": "gzip", "content-type": "application/hal+json"}
                )
                if response.status_code == 403:
                    raise ConnectionError(
                        'Connection to Komoot API failed. Please check your credentials.'
                    )
                highlight_dict = response.json()
                highlights.append(Highlight(highlight_dict))
        return highlights
